from .base.models import (
    LabelModel,
)
from .identify.models import (
    IdentifiedModel,
    Authority,
    Identifier,
)

__all__ = (
    'LabelModel',
    'IdentifiedModel',
    'Authority',
    'Identifier',
)



from rest_framework.renderers import TemplateHTMLRenderer, JSONRenderer
from django.conf import settings
from wq.db.default_settings import SRID
import re


APP_TEMPLATES = {}


def load_app_template(template_name):
    with open(template_name) as f:
        template = f.read()
    template = re.sub(
        '<title>(.+)</title>',
        '<title>{{title}}</title>',
        template
    )
    if '{{' in template:
        return template, True
    else:
        return template, False


def get_title(data, request):
    title = None
    if isinstance(data, dict):
        title = data.get('label')

    return title or settings.PROJECT_NAME


def render_app(template_name, data, request):
    if template_name not in APP_TEMPLATES:
        APP_TEMPLATES[template_name] = load_app_template(template_name)

    template, has_title = APP_TEMPLATES[template_name]
    if has_title:
        from wq.db.rest import router
        return (template
                .replace('{{title}}', get_title(data, request))
                .replace('{{base_url}}', router.get_base_url()))
    else:
        return template


class HTMLRenderer(TemplateHTMLRenderer):
    def render(self, data, accepted_media_type=None, renderer_context=None):
        if getattr(settings, 'WQ_APP_TEMPLATE', None):
            return render_app(
                settings.WQ_APP_TEMPLATE,
                data,
                (renderer_context or {}).get('request'),
            )
        return super().render(data, accepted_media_type, renderer_context)


class ESMRenderer(JSONRenderer):
    media_type = 'application/javascript'
    format = 'js'

    def render(self, data, accepted_media_type=None, renderer_context=None):
        data = super().render(data, accepted_media_type, renderer_context)
        return b"const config = " + data + b";\nexport default config;"


class GeoJSONRenderer(JSONRenderer):
    media_type = 'application/geo+json'
    format = 'geojson'

    def render(self, data, *args, **kwargs):
        if isinstance(data, list):
            features, simple = self.render_features(data)
            data = {
                'type': 'FeatureCollection',
                'features': features
            }
        elif "list" in data and isinstance(data['list'], list):
            features, simple = self.render_features(data['list'])
            data['type'] = 'FeatureCollection'
            data['features'] = features
            del data['list']

        else:
            data, simple = self.render_feature(data)

        if not simple and getattr(settings, 'SRID', SRID) != SRID:
            data['crs'] = {
                'type': 'name',
                'properties': {
                    'name': 'urn:ogc:def:crs:EPSG::%s' % settings.SRID
                }
            }
        return super().render(data, *args, **kwargs)

    def render_feature(self, obj):
        feature = {
            'type': 'Feature',
            'properties': obj
        }
        simple = False
        if 'id' in obj:
            feature['id'] = obj['id']
            del obj['id']

        if 'latitude' in obj and 'longitude' in obj:
            feature['geometry'] = {
                'type': 'Point',
                'coordinates': [obj['longitude'], obj['latitude']]
            }
            del obj['latitude']
            del obj['longitude']
            simple = True

        else:
            for key, val in list(obj.items()):
                if isinstance(val, dict) and 'type' in val and (
                        'coordinates' in val or 'geometries' in val
                        ):
                    feature['geometry'] = val
                    del obj[key]

        if 'features' in obj:
            feature['features'] = obj['features']
            feature['type'] = 'FeatureCollection'
            del obj['features']

        return feature, simple

    def render_features(self, objs):
        features = []
        has_simple = False
        for obj in objs:
            feature, simple = self.render_feature(obj)
            if simple:
                has_simple = True
                if feature['geometry']['coordinates'][0] is not None:
                    features.append(feature)
            else:
                features.append(feature)
        return features, has_simple



import os
from django.test.utils import setup_test_environment
import django
from django.core.management import call_command

os.environ.setdefault('DJANGO_SETTINGS_MODULE', "tests.settings")
setup_test_environment()
django.setup()
call_command('makemigrations', interactive=False)
call_command('migrate', interactive=False)



from wq.db.patterns import serializers as patterns

from .models import (
    CustomAttachment, CustomTypedAttachment,
    Entity, Value
    )


class CustomAttachmentSerializer(patterns.AttachmentSerializer):
    class Meta(patterns.AttachmentSerializer.Meta):
        model = CustomAttachment
        exclude = ['parent']

        object_field = 'parent'


class CustomPatternSerializer(patterns.AttachedModelSerializer):
    attachments = CustomAttachmentSerializer(many=True)


class CustomTypedAttachmentSerializer(patterns.TypedAttachmentSerializer):
    class Meta(patterns.TypedAttachmentSerializer.Meta):
        model = CustomTypedAttachment
        exclude = ['parent']

        object_field = 'parent'


class CustomTypedPatternSerializer(patterns.AttachedModelSerializer):
    attachments = CustomTypedAttachmentSerializer(many=True)


class ValueSerializerBase(patterns.TypedAttachmentSerializer):
    class Meta(patterns.TypedAttachmentSerializer.Meta):
        model = Value
        exclude = ('entity',)
        object_field = 'entity'
        type_field = 'attribute'
        type_filter = {}


class ValueSerializerCampaignID(ValueSerializerBase):
    class Meta(ValueSerializerBase.Meta):
        type_filter = {'campaign_id': '{{campaign_id}}'}


class ValueSerializerIsActiveT(ValueSerializerBase):
    class Meta(ValueSerializerBase.Meta):
        type_filter = {'is_active': '1'}


class ValueSerializerIsActiveF(ValueSerializerBase):
    class Meta(ValueSerializerBase.Meta):
        type_filter = {'is_active': '0'}


class ValueSerializerIsActiveTCampaignID(ValueSerializerBase):
    class Meta(ValueSerializerBase.Meta):
        type_filter = {'is_active': '1', 'campaign_id': '{{campaign_id}}'}


class ValueSerializerCategoryDim(ValueSerializerBase):
    class Meta(ValueSerializerBase.Meta):
        type_filter = {'category': 'dimension'}


class ValueSerializerCategoryEmpty(ValueSerializerBase):
    class Meta(ValueSerializerBase.Meta):
        type_filter = {'category': ''}


class ValueSerializerCategoryCtxt(ValueSerializerBase):
    class Meta(ValueSerializerBase.Meta):
        type_filter = {'category': '{{category}}'}


class EntitySerializerBase(patterns.AttachedModelSerializer):
    values = ValueSerializerBase(many=True)

    class Meta:
        model = Entity
        fields = '__all__'


class EntitySerializerCampaignID(EntitySerializerBase):
    values = ValueSerializerCampaignID(many=True)

    class Meta(EntitySerializerBase.Meta):
        pass


class EntitySerializerIsActiveT(EntitySerializerBase):
    values = ValueSerializerIsActiveT(many=True)

    class Meta(EntitySerializerBase.Meta):
        pass


class EntitySerializerIsActiveF(EntitySerializerBase):
    values = ValueSerializerIsActiveF(many=True)

    class Meta(EntitySerializerBase.Meta):
        pass


class EntitySerializerActiveTCampaignID(EntitySerializerBase):
    values = ValueSerializerIsActiveTCampaignID(many=True)

    class Meta(EntitySerializerBase.Meta):
        pass


class EntitySerializerCategoryDim(EntitySerializerBase):
    values = ValueSerializerCategoryDim(many=True)

    class Meta(EntitySerializerBase.Meta):
        pass


class EntitySerializerCategoryEmpty(EntitySerializerBase):
    values = ValueSerializerCategoryEmpty(many=True)

    class Meta(EntitySerializerBase.Meta):
        pass


class EntitySerializerCategoryCtxt(EntitySerializerBase):
    values = ValueSerializerCategoryCtxt(many=True)

    class Meta(EntitySerializerBase.Meta):
        pass



from .base import APITestCase
from rest_framework import status
from tests.rest_app.models import ItemType, DateModel, ChoiceModel
from tests.rest_app.serializers import (
    ChoiceLabelSerializer, DateLabelSerializer, ItemLabelSerializer,
)


class LabelTestCase(APITestCase):
    def setUp(self):
        itype = ItemType.objects.create(name="Test", pk=1)
        itype.item_set.create(name="Test 1")
        DateModel.objects.create(
            pk=1,
            name="Test",
            date="2015-01-01 12:00Z",
        )
        ChoiceModel.objects.create(
            pk=1,
            name="Test",
            choice="two",
        )

    def test_rest_date_label(self):
        response = self.client.get("/datemodels/1.json")
        self.assertTrue(status.is_success(response.status_code), response.data)
        self.assertIn('date_label', response.data)
        self.assertEqual(response.data['date_label'], "2015-01-01 06:00 AM")

    def test_rest_choice_label(self):
        response = self.client.get("/choicemodels/1.json")
        self.assertTrue(status.is_success(response.status_code), response.data)
        self.assertIn('choice_label', response.data)
        self.assertEqual(response.data['choice_label'], "Choice Two")

    def test_rest_boolean_label(self):
        response = self.client.get("/itemtypes/1.json")
        self.assertTrue(status.is_success(response.status_code), response.data)
        self.assertIn('active_label', response.data)
        self.assertEqual(response.data['active_label'], "Yes")

    def test_custom_date_label(self):
        from wq.db.rest import router

        # Default label
        item = DateModel.objects.get(pk=1)
        obj = router.serialize(item)
        self.assertEqual(obj['date_label'], '2015-01-01 06:00 AM')

        # Custom label
        obj = DateLabelSerializer(item).data
        self.assertEqual(obj['date_label'], 'January 1, 2015')

    def test_custom_choice_label(self):
        from wq.db.rest import router

        # Default label
        item = ChoiceModel.objects.get(pk=1)
        obj = router.serialize(item)
        self.assertEqual(obj['choice_label'], 'Choice Two')

        # Custom label
        obj = ChoiceLabelSerializer(item).data
        self.assertEqual(obj['choice_label'], 'TWO')

    def test_custom_fk_label(self):
        from wq.db.rest import router

        # Default labels
        item = ItemType.objects.get(pk=1).item_set.get(name='Test 1')
        obj = router.serialize(item)
        self.assertEqual(obj['type_id'], 1)
        self.assertEqual(obj['type_label'], 'Test')
        self.assertEqual(obj['label'], 'Test 1')

        # Custom labels
        obj = ItemLabelSerializer(item).data
        self.assertEqual(obj['type_id'], 'id-1')
        self.assertEqual(obj['type_label'], 'TEST')
        self.assertEqual(obj['label'], 'Test: Test 1')



from wq.db import rest
from wq.db.patterns import rest as patterns
from wq.db.patterns.identify.views import IdentifiedModelViewSet
from .models import (
    IdentifiedModel,
    CustomPatternModel, CustomTypedPatternModel, CustomType,
    Campaign, Attribute, Entity
)
from .serializers import (
    CustomPatternSerializer, CustomTypedPatternSerializer,
)
rest.router.register_model(
    IdentifiedModel,
    serializer=patterns.IdentifiedModelSerializer,
    fields='__all__',
    viewset=IdentifiedModelViewSet,
)
rest.router.register_model(
    CustomPatternModel,
    serializer=CustomPatternSerializer,
    fields='__all__',
)
rest.router.register_model(
    CustomTypedPatternModel,
    serializer=CustomTypedPatternSerializer,
    fields='__all__',
)
rest.router.register_model(
    CustomType,
    fields='__all__',
)
rest.router.register_model(
    Campaign,
    fields='__all__')
rest.router.register_model(
    Attribute,
    fields='__all__')
rest.router.register_model(
    Entity,
    fields='__all__')



from . import router
from django.utils.safestring import mark_safe
from django.urls import reverse
from django.conf import settings
from urllib.parse import quote
from html.parser import HTMLParser


def version(request):
    return {'version': router.version}


def get_base_url():
    base_url = reverse('wq:config-list').replace('/config/', '')
    if base_url != router.get_base_url():
        # FIXME: raise ImproperlyConfigured in 2.0?
        pass
    return base_url


def get_wq_path(request):
    base_url = get_base_url()
    if base_url and not request.path.startswith(base_url):
        return None
    path = request.path.replace(base_url + "/", "")
    return path


def router_info(request):
    base_url = get_base_url()
    full_path = request.path
    path = get_wq_path(request)
    if not path:
        return {}
    if request.GET:
        path += "?" + request.GET.urlencode()

    info = {
        'base_url': base_url,
        'path': path,
        'path_enc': quote(path),
        'params': request.GET,
        'full_path': full_path,
        'full_path_enc': quote(full_path),
        'prev_path': '',  # Referer?
    }

    return {
        'rt': base_url,
        'svc': base_url,
        'router_info': info,
        'pages_info': info,  # FIXME: Remove in 2.0
    }


# FIXME: Remove in 2.0
def pages_info(request):
    return router_info(request)


def wq_config(request):
    path = get_wq_path(request)
    if not path:
        return {}
    parts = path.split('/')
    user = request.user if request.user.is_authenticated else None
    wq_conf = router.get_config(user=user)
    page_conf = None
    root_conf = None

    for name, conf in wq_conf['pages'].items():
        if parts[0] == conf['url']:
            page_conf = conf
        elif conf['url'] == "":
            root_conf = conf

    if not page_conf and root_conf and len(parts) == 1:
        page_conf = root_conf

    return {
        'wq_config': wq_conf,
        'page_config': page_conf,
    }


_script_tags = None


def script_tags(request):
    global _script_tags
    if _script_tags is None:
        _script_tags = parse_script_tags()
    return {'script_tags': mark_safe(_script_tags)}


def parse_script_tags():
    filename = getattr(settings, 'WQ_SCRIPT_FILE', None)
    if not filename:
        return ''

    parser = ScriptTagParser()
    with open(filename) as f:
        parser.feed(f.read())

    return '\n'.join(parser.output)


class ScriptTagParser(HTMLParser):
    in_script = None
    output = []

    def handle_starttag(self, tag, attrs):
        if tag == 'script':
            if attrs:
                attrs = dict(attrs)
                self.output.append(
                    "<script async src='{src}'>".format(**attrs)
                )
            else:
                self.output.append("<script async>")
            self.in_script = True

    def handle_data(self, data):
        if self.in_script:
            self.output.append(data)

    def handle_endtag(self, tag):
        if tag == 'script':
            self.output.append('</script>')
            self.in_script = False



from .base import APITestCase
from rest_framework import status
from django.contrib.auth.models import User
from tests.patterns_app.models import IdentifiedModel
from wq.db.patterns.models import Authority


def ident_by_auth(idents):
    return {
        ident['authority_id']: ident for ident in idents
    }


class IdentifyTestCase(APITestCase):
    def setUp(self):
        self.auth = Authority.objects.create(
            name="Example",
            homepage="http://example.com/",
            object_url="http://example.com/pages/%s",
        )

    def test_identify_find(self):
        instance = IdentifiedModel.objects.find("Test 1")
        self.assertEqual(instance.name, "Test 1")
        self.assertEqual(instance.slug, "test-1")
        self.assertEqual(instance.primary_identifier.name, "Test 1")
        self.assertEqual(instance.primary_identifier.slug, "test-1")

        instance2 = IdentifiedModel.objects.find("Test 1")
        self.assertEqual(instance, instance2)

    def test_identify_auth_url(self):
        instance = IdentifiedModel.objects.find("Test 2")
        instance.identifiers.create(
            authority=self.auth,
            name="Test 2",
            slug="test2"
        )

        ident = instance.primary_identifier
        self.assertIsNone(ident.url)

        ident = instance.identifiers.get(authority=self.auth)
        self.assertEqual(ident.url, "http://example.com/pages/test2")

    def test_identify_order(self):
        auth2 = Authority.objects.create(
            name="Example.org",
            homepage="http://example.org/",
            object_url="http://example.org/content.php?id=%s",
        )
        instance = IdentifiedModel.objects.find("Test 3")
        instance.identifiers.create(
            authority=auth2,
            name="Test 3",
            slug="123",
        )
        instance.identifiers.create(
            authority=self.auth,
            name="Test 3",
            slug="test3"
        )
        idents = list(instance.identifiers.all())
        self.assertIsNone(idents[0].authority)
        self.assertTrue(idents[0].is_primary)
        self.assertEqual(idents[1].authority, self.auth)
        self.assertEqual(idents[2].authority, auth2)

    def test_identify_update(self):
        """
        Updating identifier slug should update object and vice-versa
        """
        instance = IdentifiedModel.objects.create(name="Test 4")
        ident = instance.primary_identifier
        self.assertIsNotNone(ident)
        self.assertEqual(ident.slug, 'test-4')
        ident.slug = 'test-4-update'
        ident.name = 'Test 4 Update'
        ident.save()
        instance = IdentifiedModel.objects.get(pk=instance.pk)
        self.assertEqual(instance.slug, 'test-4-update')
        self.assertEqual(instance.name, 'Test 4 Update')
        instance.slug = "test-4-update-2"
        instance.name = "Test 4 Update 2"
        instance.save()
        self.assertEqual(instance.primary_identifier.slug, 'test-4-update-2')
        self.assertEqual(instance.primary_identifier.name, 'Test 4 Update 2')

    def test_identify_autocreate(self):
        """
        Deleting identifier and re-saving should create a new one.
        """
        instance = IdentifiedModel.objects.create(
            name="Test 1",
        )

        instance.identifiers.all().delete()
        self.assertIsNone(instance.primary_identifier)
        instance.save()
        self.assertIsNotNone(instance.primary_identifier)

        self.assertEqual(instance.primary_identifier.slug, "test-1")


class IdentifyRestTestCase(APITestCase):
    def setUp(self):
        self.user = User.objects.create(username='testuser', is_superuser=True)
        self.auth = Authority.objects.create(
            name="Example",
            homepage="http://example.com/",
            object_url="http://example.com/pages/%s",
        )
        self.instance = IdentifiedModel.objects.find("Test 1")
        self.instance.identifiers.create(
            authority=self.auth,
            name="Test 1",
            slug="test1"
        )
        self.client.force_authenticate(user=self.user)

    def test_identify_config(self):
        response = self.client.get('/config.json')
        self.maxDiff = None
        self.assertEqual([
            {
                'name': 'name',
                'label': 'Name',
                'type': 'string',
                'wq:length': 255,
                'bind': {'required': True},
            }, {
                'name': 'slug',
                'label': 'Slug',
                'type': 'string',
            }, {
                'name': 'identifiers',
                'label': 'Identifiers',
                'type': 'repeat',
                'bind': {'required': True},
                'children': [{
                    'name': 'name',
                    'label': 'Name',
                    'type': 'string',
                    'bind': {'required': True},
                    'wq:length': 255,
                }, {
                    'name': 'slug',
                    'label': 'Slug',
                    'type': 'string',
                    'wq:length': 50,
                }, {
                    'name': 'authority',
                    'label': 'Authority',
                    'type': 'select one',
                    'wq:ForeignKey': 'authority',
                }, {
                    'name': 'is_primary',
                    'label': 'Is Primary',
                    'type': 'select one',
                    'choices': [
                        {'name': True, 'label': 'Yes'},
                        {'name': False, 'label': 'No'},
                    ],
                }],
                'initial': {
                    'type_field': 'authority',
                    'filter': {},
                }
            },
        ], response.data['pages']['identifiedmodel']['form'])

    def test_identify_detail_nested_identifiers(self):
        response = self.client.get('/identifiedmodels/test-1.json')
        self.assertIn('identifiers', response.data)
        self.assertEqual(response.data['identifiers'][0]['slug'], 'test-1')

    def test_identify_list_nested_identifiers(self):
        response = self.client.get('/identifiedmodels.json')
        self.assertIn('identifiers', response.data['list'][0])

    def test_identify_post(self):
        form = {
            'name': 'Test 2',

            'identifiers[0][name]': "Test 2",

            'identifiers[1][name]': "Test 2",
            'identifiers[1][authority_id]': self.auth.pk,
            'identifiers[1][slug]': "test2",
        }

        response = self.client.post('/identifiedmodels.json', form)
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.data
        )
        self.assertEqual(response.data['name'], "Test 2")
        self.assertEqual(response.data['id'], "test-2")

        self.assertIn("identifiers", response.data)
        self.assertEqual(len(response.data["identifiers"]), 2)

        idents = ident_by_auth(response.data["identifiers"])

        self.assertEqual(idents[None]["slug"], "test-2")
        self.assertEqual(idents[self.auth.pk]["slug"], "test2")

    def test_identify_put(self):
        ident = self.instance.identifiers.get(authority=self.auth)
        form = {
            'name': 'Test 1 - Updated',

            'identifiers[0][id]': self.instance.primary_identifier.pk,
            'identifiers[0][name]': "Test 1 - Updated",
            'identifiers[0][slug]': "test-1-updated",

            'identifiers[1][id]': ident.pk,
            'identifiers[1][name]': "Test 1B",
            'identifiers[1][slug]': "test1b",
            'identifiers[1][type]': self.auth.pk,
        }
        url = (
            '/identifiedmodels/%s.json' % self.instance.slug
        )

        response = self.client.put(url, form)
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.data
        )
        self.instance = IdentifiedModel.objects.get(pk=self.instance.pk)
        self.assertEqual(self.instance.name, "Test 1 - Updated")
        self.assertEqual(response.data['id'], "test-1-updated")

        self.assertIn("identifiers", response.data)
        self.assertEqual(len(response.data["identifiers"]), 2)

        idents = ident_by_auth(response.data["identifiers"])

        self.assertEqual(idents[None]["slug"], "test-1-updated")
        self.assertEqual(idents[self.auth.pk]["slug"], "test1b")
        self.assertEqual(
            idents[self.auth.pk]["url"], "http://example.com/pages/test1b"
        )

    def test_identify_post_duplicate_auto_slug(self):
        form = {
            'name': 'Test 3',
        }
        response = self.client.post('/identifiedmodels.json', form)
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.data
        )
        self.assertEqual(response.data['id'], 'test-3')

        response = self.client.post('/identifiedmodels.json', form)
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.data
        )
        self.assertEqual(response.data['id'], 'test-3-1')

    def test_identify_post_duplicate_auto_ident_slug(self):
        form = {
            'name': 'Test 4',
            'identifiers[0][name]': 'Test 4 Ident',
        }
        response = self.client.post('/identifiedmodels.json', form)
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.data
        )
        self.assertEqual(response.data['id'], 'test-4-ident', response.data)

        response = self.client.post('/identifiedmodels.json', form)
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.data
        )
        self.assertEqual(response.data['id'], 'test-4-ident-1')

    def test_identify_post_duplicate_explicit_slug(self):
        form = {
            'name': 'Test 5',
            'slug': 'test-5',
        }
        response = self.client.post('/identifiedmodels.json', form)
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.data
        )

        response = self.client.post('/identifiedmodels.json', form)
        self.assertEqual(
            response.status_code, status.HTTP_400_BAD_REQUEST, response.data
        )
        self.assertEqual(
            response.data['non_field_errors'],
            ["The fields slug must make a unique set."],
        )

    def test_identify_post_duplicate_explicit_ident_slug(self):
        # FIXME: Handle this case
        return
        form = {
            'name': 'Test 6',
            'identifiers[0][name]': 'Test 6',
            'identifiers[0][slug]': 'test-6',
        }
        response = self.client.post('/identifiedmodels.json', form)
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.data
        )

        response = self.client.post('/identifiedmodels.json', form)
        self.assertEqual(
            response.status_code, status.HTTP_400_BAD_REQUEST, response.data
        )

    def test_identify_alt_ident(self):
        form = {
            'name': 'Test 7',
            'identifiers[0][name]': "Test 7",

            'identifiers[1][name]': "Test 7 Alt",
            'identifiers[1][authority_id]': self.auth.pk,
        }

        response = self.client.post('/identifiedmodels.json', form)
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.data
        )

        response = self.client.get('/identifiedmodels/test-7-alt.json', form)
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.data
        )
        self.assertEqual(response.data['id'], 'test-7')

    def test_filter_backend(self):
        self.instance.filterablemodel_set.create(
            pk=1,
            name="Filter 1",
        )
        self.instance.filterablemodel_set.create(
            pk=2,
            name="Filter 2",
        )
        instance2 = IdentifiedModel.objects.find("Test 2")
        instance2.filterablemodel_set.create(
            pk=3,
            name="Filter 3",
        )
        instance3 = IdentifiedModel.objects.find("Test 3")
        instance3.filterablemodel_set.create(
            pk=4,
            name="Filter 4",
        )

        response = self.client.get("/filterable/test1/test-3?format=json")
        self.assertEqual(
            status.HTTP_200_OK, response.status_code
        )
        self.assertIn('list', response.data)
        self.assertEqual(3, len(response.data['list']))
        self.assertEqual(1, response.data['list'][0]['id'])
        self.assertEqual(2, response.data['list'][1]['id'])
        self.assertEqual(4, response.data['list'][2]['id'])



import unittest
from .base import APITestCase
import json
from django.conf import settings


class ConfigTestCase(APITestCase):
    def get_config(self, page_name):
        response = self.client.get('/config.json')
        result = json.loads(response.content.decode('utf-8'))
        self.assertIn('pages', result)
        self.assertIn(page_name, result['pages'])
        return result['pages'][page_name]

    def get_field(self, page_config, field_name, allow_none=False):
        self.assertIn('form', page_config)
        for field in page_config['form']:
            if field['name'] == field_name:
                return field
        if not allow_none:
            self.fail("Could not find %s" % field_name)

    # Test existence and content of config.json
    def test_rest_config_json(self):
        response = self.client.get('/config.json')
        result = json.loads(response.content.decode('utf-8'))
        self.assertIn("pages", result)

        # Extra config
        self.assertIn("debug", result)

    def test_rest_index_json(self):
        from wq.db.rest import router
        result = router.get_index(None)
        self.assertIn("pages", result)
        self.assertIn("list", result["pages"][0])
        self.assertNotIn("list", result["pages"][-1])

    def test_rest_config_meta(self):
        conf = self.get_config('item')
        self.assertEqual({
            'name': 'item',
            'url': 'items',
            'list': True,
            'cache': 'none',
            'form': conf['form'],
            'label_template': '{{name}}',
            'verbose_name': 'item',
            'verbose_name_plural': 'items',
            'ordering': ['type', 'name'],
        }, conf)

    def test_rest_config_json_fields(self):
        self.assertEqual([
            {
                'name': 'name',
                'label': 'Name',
                'type': 'string',
                'wq:length': 10,
                'bind': {'required': True},
            },
            {
                'name': 'date',
                'label': 'Date',
                'type': 'dateTime',
                'bind': {'required': True},
            },
            {
                'name': 'empty_date',
                'label': 'Empty Date',
                'type': 'dateTime',
            },
        ], self.get_config('datemodel')['form'])

    def test_rest_config_json_choices(self):
        conf = self.get_config('choicemodel')
        self.assertEqual([
            {
                'name': 'name',
                'label': 'Name',
                'hint': 'Enter Name',
                'type': 'string',
                'wq:length': 10,
                'bind': {'required': True},
            },
            {
                'name': 'choice',
                'label': 'Choice',
                'hint': 'Pick One',
                'type': 'select one',
                'bind': {'required': True},
                'choices': [{
                    'name': 'one',
                    'label': 'Choice One',
                }, {
                    'name': 'two',
                    'label': 'Choice Two',
                }, {
                    'name': 'three',
                    'label': 'Choice Three',
                }]
            },
            {
                'name': 'grouped_choice',
                'label': 'Grouped Choice',
                'hint': 'Pick One',
                'type': 'select one',
                'bind': {'required': True},
                'choices': [{
                    'name': 'one',
                    'label': 'Choice One',
                    'group': 'Group 1',
                }, {
                    'name': 'two',
                    'label': 'Choice Two',
                    'group': 'Group 1',
                }, {
                    'name': 'three',
                    'label': 'Choice Three',
                    'group': 'Group 1',
                }]
            },
        ], conf['form'])

    def test_rest_config_json_boolean_choices(self):
        conf = self.get_config('itemtype')
        self.assertEqual([
            {
                'name': 'name',
                'label': 'Name',
                'type': 'string',
                'wq:length': 10,
                'bind': {'required': True},
            },
            {
                'name': 'active',
                'label': 'Active',
                'type': 'select one',
                'choices': [{
                    'name': True,
                    'label': 'Yes',
                }, {
                    'name': False,
                    'label': 'No',
                }],
            }
        ], conf['form'])

    def test_rest_config_json_rels(self):
        pconf = self.get_config('parent')
        self.assertEqual({
            'name': 'children',
            'label': 'Children',
            'type': 'repeat',
            'bind': {'required': True},
            'children': [{
                'name': 'name',
                'label': 'Name',
                'type': 'string',
                'wq:length': 10,
                'bind': {'required': True},
            }]
        }, self.get_field(pconf, 'children'))

        cconf = self.get_config('child')
        self.assertEqual({
            'name': 'parent',
            'label': 'Parent',
            'type': 'select one',
            'wq:ForeignKey': 'parent',
            'bind': {'required': True},
        }, self.get_field(cconf, 'parent'))

    def test_rest_config_json_override(self):
        iconf = self.get_config('item')
        self.assertEqual({
            'name': 'type',
            'label': 'Type',
            'type': 'select one',
            'wq:ForeignKey': 'itemtype',
            'filter': {'active':  ['1', '{{#id}}0{{/id}}{{^id}}1{{/id}}']},
            'bind': {'required': True},
        }, self.get_field(iconf, 'type'))

    def test_rest_config_json_label(self):
        pconf = self.get_config('slugrefparent')
        self.assertIn('label_template', pconf)
        self.assertEqual(
            "{{name}}{{#ref_id}} ({{ref}}){{/ref_id}}",
            pconf['label_template'],
        )

    @unittest.skipUnless(settings.WITH_GIS, "requires GIS")
    def test_rest_config_subtype_gis(self):
        conf = self.get_config('geometrymodel')
        field = self.get_field(conf, 'geometry')
        self.assertEqual('geoshape', field['type'])

        conf = self.get_config('pointmodel')
        field = self.get_field(conf, 'geometry')
        self.assertEqual('geopoint', field['type'])

    def test_rest_config_subtype(self):
        conf = self.get_config('filemodel')
        field = self.get_field(conf, 'file')
        self.assertEqual('file', field['type'])

        conf = self.get_config('imagemodel')
        field = self.get_field(conf, 'image')
        self.assertEqual('image', field['type'])

        conf = self.get_config('item')
        field = self.get_field(conf, 'name')
        self.assertEqual('string', field['type'])

        conf = self.get_config('rootmodel')
        field = self.get_field(conf, 'description')
        self.assertEqual('text', field['type'])

    def test_rest_config_field_order(self):
        conf = self.get_config('slugrefparent')
        self.assertEqual(conf['form'][0]['name'], 'ref')
        self.assertEqual(conf['form'][1]['name'], 'name')

    def test_rest_list_exclude_config(self):
        conf = self.get_config('expensivemodel')
        self.get_field(conf, 'name')
        self.get_field(conf, 'expensive')
        self.assertIsNone(
            self.get_field(conf, 'more_expensive', allow_none=True)
        )

    def test_rest_virtual_fieldset(self):
        conf = self.get_config('fieldsetmodel')
        self.assertEqual([
            {
                'name': 'general',
                'type': 'group',
                'label': 'General',
                'children': [
                    {
                        'name': 'name',
                        'label': 'Name',
                        'type': 'string',
                        'wq:length': 50,
                        'bind': {'required': True},
                    },
                    {
                        'name': 'title',
                        'label': 'Title',
                        'type': 'string',
                        'wq:length': 20,
                        'bind': {'required': True},
                    },
                    {
                        'name': 'status',
                        'label': 'Status',
                        'type': 'string',
                        'disabled': True,
                    }
                ]
            },
            {
                'name': 'contact',
                'type': 'group',
                'label': 'Contact Information',
                'control': {'appearance': 'contact-fieldset'},
                'children': [
                    {
                        'name': 'address',
                        'label': 'Address',
                        'type': 'string',
                        'wq:length': 255,
                        'bind': {'required': True},
                    },
                    {
                        'name': 'city',
                        'label': 'City',
                        'type': 'string',
                        'wq:length': 255,
                        'bind': {'required': True},
                    },
                ]
            },
        ], conf['form'])



from wq.db.rest.serializers import ModelSerializer
from rest_framework import serializers
from natural_keys import NaturalKeySerializer, NaturalKeyModelSerializer
from django.conf import settings


class AttachmentListSerializer(serializers.ListSerializer):
    def to_representation(self, data):
        data = super(AttachmentListSerializer, self).to_representation(data)
        if not self.parent:
            return

        wq_config = self.child.get_wq_config()
        initial = wq_config.get('initial', None)
        if initial and not data and self.parent.is_detail:
            if self.parent.instance and not self.parent.instance.pk:
                data = self.default_attachments(initial)

        if not getattr(settings, 'WQ_APP_TEMPLATE', None):
            # FIXME: remove in 2.0
            for i, row in enumerate(data):
                row['@index'] = i

        return data

    def default_attachments(self, initial):
        data = []
        for i in range(0, int(initial)):
            data.append({'new_attachment': True})
        return data


class TypedAttachmentListSerializer(AttachmentListSerializer):
    def check_empty(self, value):
        if value is None or value == '':
            return True
        else:
            return False

    def get_value(self, dictionary):
        # Handle attachments that are submitted together with their parent
        value = super(TypedAttachmentListSerializer, self).get_value(
            dictionary
        )
        if not isinstance(value, list):
            return []
        for i, row in enumerate(value):
            empty = True
            if isinstance(row, dict):
                for key, val in row.items():
                    if key in ('@index', self.child.Meta.type_field):
                        continue
                    elif not self.check_empty(val):
                        empty = False
            if empty:
                value[i] = None
        return value

    def default_attachments(self, initial):
        if not isinstance(initial, dict):
            return super(
                TypedAttachmentListSerializer, self
            ).add_attachments(initial)

        wq_config = self.child.get_wq_config()
        type_field = None
        for field in wq_config['form']:
            if field['name'] == initial.get('type_field'):
                type_field = field.copy()

        if not type_field:
            return []

        if initial.get('filter'):
            type_field['filter'] = initial['filter']
        types = self.child.get_lookup_choices(
            type_field,
            self.context.get('request').GET.dict()
            )

        data = [{
           'new_attachment': True,
           type_field['name'] + '_id': row['id'],
           type_field['name'] + '_label': row['label'],
        } for row in types]

        return data


class AttachmentSerializer(ModelSerializer):
    id = serializers.IntegerField(required=False)

    def __init__(self, *args, **kwargs):
        kwargs['allow_null'] = True
        super(AttachmentSerializer, self).__init__(*args, **kwargs)

    class Meta:
        list_serializer_class = AttachmentListSerializer


class TypedAttachmentSerializer(AttachmentSerializer):
    def get_wq_config(self):
        config = super(TypedAttachmentSerializer, self).get_wq_config()
        if 'initial' not in config:
            config['initial'] = {
                'type_field': self.Meta.type_field.replace('_id', ''),
                'filter': self.Meta.type_filter,
            }
        return config

    class Meta:
        # Don't validate these fields (items are saved with their parent)
        exclude = ("content_type", "object_id",)
        list_serializer_class = TypedAttachmentListSerializer

        # patterns-specific meta
        type_field = 'type_id'
        type_filter = {}
        object_field = 'content_object'


class AttachedModelSerializer(ModelSerializer):
    def create(self, validated_data):
        model_data, attachment_data = self.extract_attachments(validated_data)
        instance = super(AttachedModelSerializer, self).create(model_data)

        fields = self.get_fields()
        for name in attachment_data:
            model = fields[name].child.Meta.model
            for attachment in attachment_data[name]:
                if not attachment:
                    continue
                self.set_parent_object(attachment, instance, name)
                self.create_attachment(model, attachment, name)
        return instance

    def update(self, instance, validated_data):
        model_data, attachment_data = self.extract_attachments(validated_data)
        obj = super(
            AttachedModelSerializer, self
        ).update(instance, model_data)

        fields = self.get_fields()
        for name in attachment_data:
            model = fields[name].child.Meta.model
            for attachment in attachment_data[name]:
                if not attachment:
                    continue
                self.set_parent_object(attachment, instance, name)
                if 'id' in attachment:
                    exist = self.get_attachment(model, attachment['id'])
                    self.update_attachment(exist, attachment, name)
                else:
                    self.create_attachment(model, attachment, name)
        return obj

    def extract_attachments(self, validated_data):
        fields = self.get_fields()
        attachment_data = {}
        for name, field in fields.items():
            if isinstance(field, AttachmentListSerializer):
                attachment_data[name] = validated_data.pop(name, [])
        return validated_data, attachment_data

    def set_parent_object(self, attachment, instance, name):
        serializer = self.get_fields()[name].child
        fk_name = getattr(serializer.Meta, 'object_field', 'content_object')
        attachment[fk_name] = instance

    def get_attachment(self, model, pk):
        return model.objects.get(pk=pk)

    def update_attachment(self, exist, attachment, name):
        field = self.get_fields()[name]
        attachment.pop('id')
        field.child.update(exist, attachment)

    def create_attachment(self, model, attachment, name):
        field = self.get_fields()[name]
        field.child.create(attachment)


class NaturalKeyModelSerializer(NaturalKeyModelSerializer, ModelSerializer):
    def get_fields(self):
        fields = ModelSerializer.get_fields(self)
        fields.update(self.build_natural_key_fields())
        return fields

    def get_wq_field_info(self, name, field, model=None):
        if isinstance(field, NaturalKeySerializer):
            children = [
                self.get_wq_field_info(n, f, model=field.Meta.model)
                for n, f in field.get_fields().items()
            ]
            if len(children) == 1:
                info = children[0]
                info['name'] = name + '[%s]' % info['name']

                fk = self.get_wq_foreignkey_info(field.Meta.model)
                if fk:
                    info['wq:ForeignKey'] = fk
                    info['type'] = 'select one'
            else:
                info = {
                    'name': name,
                    'type': 'group',
                    'bind': {'required': True},
                    'children': children
                }
            info['label'] = field.label or name.replace('_', ' ').title()
            return info
        else:
            return super(NaturalKeyModelSerializer, self).get_wq_field_info(
                name, field, model
            )


class NaturalKeyAttachedModelSerializer(NaturalKeyModelSerializer,
                                        AttachedModelSerializer):

    def create(self, validated_data):
        self.convert_natural_keys(
            validated_data
        )
        return AttachedModelSerializer.create(
            self, validated_data
        )



from rest_framework import serializers
from wq.db.rest.serializers import ModelSerializer
from .models import (
    OneToOneModel, ExtraModel, Child, ChoiceModel, DateModel, Item,
    ExpensiveModel, FieldsetModel
)


class RootModelSerializer(ModelSerializer):
    extramodels = ModelSerializer.for_model(
        ExtraModel, include_fields="__all__"
    )(many=True)
    onetoonemodel = ModelSerializer.for_model(
        OneToOneModel, include_fields="__all__"
    )()


class ChildSerializer(ModelSerializer):
    class Meta:
        model = Child
        exclude = ('parent',)


class ParentSerializer(ModelSerializer):
    children = ChildSerializer(many=True)


class ItemSerializer(ModelSerializer):
    class Meta:
        wq_field_config = {
            'type': {
                 'filter': {
                     'active': [
                         '1',

                         # Allow inactive types when editing existing items
                         '{{#id}}0{{/id}}{{^id}}1{{/id}}',
                     ]
                 }
            }
        }


class SlugRefChildSerializer(ModelSerializer):
    class Meta:
        wq_field_config = {
            'parent': {
                 'filter': {
                     'ref_id': 'test'
                 }
            }
        }


class ChoiceLabelSerializer(ModelSerializer):
    choice_label = serializers.SerializerMethodField()

    def get_choice_label(self, instance):
        return instance.choice.upper()

    class Meta:
        model = ChoiceModel
        fields = "__all__"


class DateLabelSerializer(ModelSerializer):
    date_label = serializers.SerializerMethodField()

    def get_date_label(self, instance):
        return instance.date.strftime('%B %-d, %Y')

    class Meta:
        model = DateModel
        fields = "__all__"


class ItemLabelSerializer(ModelSerializer):
    label = serializers.SerializerMethodField()
    type_id = serializers.SerializerMethodField()
    type_label = serializers.SerializerMethodField()

    def get_label(self, instance):
        return '%s: %s' % (instance.type.name, instance.name)

    def get_type_id(self, instance):
        return 'id-%s' % instance.type_id

    def get_type_label(self, instance):
        return instance.type.name.upper()

    class Meta:
        model = Item
        fields = "__all__"


class ExpensiveSerializer(ModelSerializer):
    class Meta:
        fields = "__all__"
        list_exclude = ('expensive', 'more_expensive')
        config_exclude = ('more_expensive',)
        model = ExpensiveModel


class FieldsetSerializer(ModelSerializer):
    status = serializers.SerializerMethodField(
        style={'wq_config': {'disabled': True}}
    )
    is_complete = serializers.SerializerMethodField()

    def get_status(self, instance):
        return 'OK'

    def get_is_complete(self, instance):
        return True

    class Meta:
        model = FieldsetModel
        fields = "__all__"
        wq_fieldsets = {
            'general': {
                'label': 'General',
                'fields': ['name', 'title', 'status']
            },
            'contact': {
                'label': 'Contact Information',
                'control': {'appearance': 'contact-fieldset'},
                'fields': ['address', 'city']
            }
        }



from wq.db import rest
from .base import APITestCase
from rest_framework import status
from django.contrib.auth.models import User
from tests.patterns_app.models import (
    CustomPatternModel, CustomTypedPatternModel, CustomType,
    Campaign, Attribute, Entity
    )
from tests.patterns_app.serializers import (
    EntitySerializerBase, EntitySerializerCampaignID,
    EntitySerializerIsActiveT, EntitySerializerIsActiveF,
    EntitySerializerActiveTCampaignID, EntitySerializerCategoryDim,
    EntitySerializerCategoryEmpty, EntitySerializerCategoryCtxt,
    )


class CustomPatternTestCase(APITestCase):
    def setUp(self):
        self.user = User.objects.create(username='testuser', is_superuser=True)
        self.client.force_authenticate(user=self.user)
        self.type = CustomType.objects.create(
            name="Custom",
        )
        self.instance = CustomPatternModel.objects.create(
            name="Test 1",
        )
        self.typeinstance = CustomTypedPatternModel.objects.create(
            name="Test 2",
        )
        self.campaign1 = Campaign.objects.create()
        self.campaign2 = Campaign.objects.create()
        self.att1 = Attribute.objects.create(
            name='Width',
            is_active=True,
            campaign=self.campaign1,
            category='dimension'
            )
        self.att2 = Attribute.objects.create(
            name='Height',
            is_active=False,
            campaign=self.campaign1,
            category='dimension'
            )
        self.att3 = Attribute.objects.create(
            name='Color',
            is_active=True,
            campaign=self.campaign2,
            category='color'
            )
        self.att4 = Attribute.objects.create(
            name='Size',
            is_active=False,
            campaign=self.campaign2,
            category=''
            )

    def test_customtypedpattern_config(self):
        response = self.client.get('/config.json')
        self.assertEqual([
            {
                'name': 'name',
                'label': 'Name',
                'type': 'string',
                'bind': {'required': True},
                'wq:length': 10,
            }, {
                'name': 'attachments',
                'label': 'Attachments',
                'type': 'repeat',
                'bind': {'required': True},
                'children': [{
                    'name': 'name',
                    'label': 'Name',
                    'type': 'string',
                    'wq:length': 10,
                }, {
                    'name': 'value',
                    'label': 'Value',
                    'type': 'decimal',
                }, {
                    'name': 'type',
                    'label': 'Type',
                    'type': 'select one',
                    'wq:ForeignKey': 'customtype',
                    'bind': {'required': True},
                }],
                'initial': {
                    'type_field': 'type',
                    'filter': {}
                }
            },
        ], response.data['pages']['customtypedpatternmodel']['form'])

    def test_custompattern_post(self):
        form = {
            'name': 'Test 3',

            'attachments[0][name]': 'Attachment',
            'attachments[1][name]': 'Attachment',
        }

        response = self.client.post('/custompatternmodels.json', form)
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.data
        )
        self.assertEqual(response.data['name'], "Test 3")
        self.assertIn("attachments", response.data)
        atts = response.data['attachments']
        self.assertEqual(len(atts), 2)
        self.assertEqual(atts[0]['name'], 'Attachment')

    def test_custompattern_put(self):
        form = {
            'name': "Test 1'",
            'attachments[0][name]': "Attachment"
        }

        response = self.client.put(
            '/custompatternmodels/%s.json' % self.instance.pk,
            form
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.data
        )
        self.instance = CustomPatternModel.objects.get(pk=self.instance.pk)
        self.assertEqual(self.instance.name, "Test 1'")
        self.assertEqual(self.instance.attachments.count(), 1)
        self.assertEqual(self.instance.attachments.first().name, "Attachment")

    def test_customtypedpattern_post(self):
        form = {
            'name': 'Test 4',
            'attachments[0][name]': 'Attachment',
            'attachments[0][type_id]': self.type.pk,
        }

        response = self.client.post('/customtypedpatternmodels.json', form)
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.data
        )
        self.assertEqual(response.data['name'], "Test 4")
        self.assertIn("attachments", response.data)
        atts = response.data['attachments']
        self.assertEqual(len(atts), 1)
        self.assertEqual(atts[0]['type_label'], 'Custom')

    def test_customtypedpattern_post_empty(self):
        form = {
            'name': 'Test 5',
            'attachments': [
                 {'value': '', 'type_id': self.type.pk, 'name': ''},
                 {'value': 0,  'type_id': self.type.pk, 'name': ''},
            ]
        }

        response = self.client.post(
            '/customtypedpatternmodels.json', form, format='json'
        )
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.data
        )
        self.assertIn("attachments", response.data)
        atts = response.data['attachments']
        self.assertEqual(len(atts), 1)
        self.assertEqual(atts[0]['type_label'], 'Custom')
        self.assertEqual(atts[0]['value'], 0)

    def test_customtypedpattern_put(self):
        form = {
            'name': "Test 2'",
            'attachments[0][name]': "Attachment",
            'attachments[0][type_id]': self.type.pk,
        }

        response = self.client.put(
            '/customtypedpatternmodels/%s.json' % self.typeinstance.pk,
            form
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.data
        )
        self.typeinstance = CustomTypedPatternModel.objects.get(
            pk=self.typeinstance.pk
        )
        self.assertEqual(self.typeinstance.name, "Test 2'")
        self.assertEqual(self.typeinstance.attachments.count(), 1)
        self.assertEqual(
            self.typeinstance.attachments.first().type, self.type
        )

    def test_eavfilter_empty(self):
        rest.router.register_serializer(
            Entity,
            EntitySerializerBase)

        response = self.client.get('/entities/new.json')
        self.assertEqual(
            [self.att1.id, self.att2.id, self.att3.id, self.att4.id],
            [i['attribute_id'] for i in response.data['values']])

    def test_eavfilter_cid(self):
        rest.router.register_serializer(
            Entity,
            EntitySerializerCampaignID)

        response = self.client.get(
            '/entities/new.json?campaign_id={}'.format(self.campaign2.id))
        self.assertEqual(
            [self.att3.id, self.att4.id],
            [i['attribute_id'] for i in response.data['values']])

    def test_eavfilter_cid_miss(self):
        rest.router.register_serializer(
            Entity,
            EntitySerializerCampaignID)

        response = self.client.get('/entities/new.json?foo=bar')
        self.assertEqual(
            [],
            [i['attribute_id'] for i in response.data['values']])

    def test_eavfilter_isactive(self):
        rest.router.register_serializer(
            Entity,
            EntitySerializerIsActiveT)

        response = self.client.get('/entities/new.json')
        self.assertEqual(
            [self.att1.id, self.att3.id],
            [i['attribute_id'] for i in response.data['values']])

        rest.router.register_serializer(
            Entity,
            EntitySerializerIsActiveF)

        response = self.client.get('/entities/new.json')
        self.assertEqual(
            [self.att2.id, self.att4.id],
            [i['attribute_id'] for i in response.data['values']])

    def test_eavfilter_isactivecid(self):
        rest.router.register_serializer(
            Entity,
            EntitySerializerActiveTCampaignID)

        response = self.client.get(
            '/entities/new.json?campaign_id={}'.format(self.campaign1.id))
        self.assertEqual(
            [self.att1.id],
            [i['attribute_id'] for i in response.data['values']])

    def test_eavfilter_category(self):
        rest.router.register_serializer(
            Entity,
            EntitySerializerCategoryDim)

        response = self.client.get('/entities/new.json')
        self.assertEqual(
            [self.att1.id, self.att2.id],
            [i['attribute_id'] for i in response.data['values']])

    def test_eavfilter_category_empty(self):
        rest.router.register_serializer(
            Entity,
            EntitySerializerCategoryEmpty)

        response = self.client.get('/entities/new.json')
        self.assertEqual(
            [self.att4.id],
            [i['attribute_id'] for i in response.data['values']])

    def test_eavfilter_category_ctxt_empty(self):
        rest.router.register_serializer(
            Entity,
            EntitySerializerCategoryCtxt)

        response = self.client.get('/entities/new.json?category=')
        self.assertEqual(
            [],
            [i['attribute_id'] for i in response.data['values']])



from .base import APITestCase
from rest_framework import status
from tests.rest_app.models import (
    RootModel, OneToOneModel, ForeignKeyModel, ExtraModel, UserManagedModel,
    Parent, ItemType, SlugModel,
)
from django.contrib.auth.models import User


class RestTestCase(APITestCase):
    def setUp(self):
        instance = RootModel.objects.create(
            slug='instance',
            description="Test",
        )
        for cls in OneToOneModel, ForeignKeyModel, ExtraModel:
            cls.objects.create(
                root=instance,
            )
        self.user = User.objects.create(username="testuser")
        UserManagedModel.objects.create(id=1, user=self.user)
        parent = Parent.objects.create(name="Test", pk=1)
        parent.children.create(name="Test 1")
        parent.children.create(name="Test 2")
        parent2 = Parent.objects.create(name="Test 2", pk=2)
        parent2.children.create(name="Test 1")
        itype = ItemType.objects.create(name="Test", pk=1)
        itype.item_set.create(name="Test 1")
        itype.item_set.create(name="Test 2")
        SlugModel.objects.create(
            code="test",
            name="Test",
        )

    # Test url="" use case
    def test_rest_list_at_root(self):
        response = self.client.get("/.json")
        self.assertTrue(status.is_success(response.status_code), response.data)
        self.assertTrue(len(response.data['list']) == 1)

    def test_rest_detail_at_root(self):
        response = self.client.get('/instance.json')
        self.assertTrue(status.is_success(response.status_code), response.data)
        self.assertTrue(response.data['description'] == "Test")

    # Test nested models with foreign keys
    def test_rest_detail_nested_foreignkeys(self):
        response = self.client.get('/instance.json')

        # Include explicitly declared serializers for related fields
        self.assertIn("onetoonemodel", response.data)
        self.assertEqual(
            response.data["onetoonemodel"]["label"],
            "onetoonemodel for instance"
        )
        self.assertIn("extramodels", response.data)
        self.assertEqual(
            response.data["extramodels"][0]["label"],
            "extramodel for instance"
        )

        # Related fields without explicit serializers will not be included
        self.assertNotIn("extramodel_set", response.data)
        self.assertNotIn("foreignkeymodel_set", response.data)
        self.assertNotIn("foreignkeymodels", response.data)

    def test_rest_filter_by_parent(self):
        response = self.client.get('/parents/1/children.json')
        self.assertIn("list", response.data)
        self.assertEqual(len(response.data['list']), 2)

        response = self.client.get('/itemtypes/1/items.json')
        self.assertIn("list", response.data)
        self.assertEqual(len(response.data['list']), 2)

    def test_rest_target_to_children(self):
        response = self.client.get('/children-by-parents.json')
        self.assertIn("list", response.data)
        self.assertEqual(len(response.data['list']), 2)
        self.assertIn("target", response.data)
        self.assertEqual(response.data['target'], 'children')

    def test_rest_detail_user_serializer(self):
        response = self.client.get('/usermanagedmodels/1.json')
        self.assertIn('user', response.data)
        self.assertIn('label', response.data['user'])
        self.assertNotIn('password', response.data['user'])

    def test_rest_multi(self):
        lists = ['usermanagedmodels', 'items', 'children']
        response = self.client.get(
            "/multi.json?lists=" + ",".join(lists)
        )
        for listurl in lists:
            self.assertIn(listurl, response.data)
            self.assertIn("list", response.data[listurl])
            self.assertGreater(len(response.data[listurl]["list"]), 0)

    def test_rest_custom_lookup(self):
        response = self.client.get('/slugmodels/test.json')
        self.assertTrue(status.is_success(response.status_code), response.data)
        self.assertEqual(response.data['id'], 'test')

    def test_rest_default_per_page(self):
        response = self.client.get('/parents.json')
        self.assertTrue(status.is_success(response.status_code), response.data)
        self.assertEqual(response.data['per_page'], 50)

    def test_rest_custom_per_page(self):
        response = self.client.get('/children.json')
        self.assertTrue(status.is_success(response.status_code), response.data)
        self.assertEqual(response.data['per_page'], 100)

    def test_rest_limit(self):
        response = self.client.get('/children.json?limit=10')
        self.assertTrue(status.is_success(response.status_code), response.data)
        self.assertEqual(response.data['per_page'], 10)

    def test_rest_cache_all(self):
        for num in range(2, 101):
            ItemType.objects.create(
                pk=num,
                name='Type #%s' % num
            )

        tests = [
             (100, '/itemtypes.json'),
             (50, '/itemtypes/'),
             (50, '/itemtypes.json?page=1'),
        ]
        for expect_items, url in tests:
            response = self.client.get(url)
            self.assertTrue(
                status.is_success(response.status_code), response.data
            )
            self.assertEqual(
                len(response.data['list']), expect_items,
                "%s should return %s items" % (url, expect_items)
            )
            self.assertEqual(response.data['pages'], 2)
            self.assertEqual(response.data['per_page'], 50)
            self.assertEqual(response.data['count'], 100)

    def test_rest_cache_filter(self):
        other_user = User.objects.create(username='otheruser')
        UserManagedModel.objects.create(id=2, user=other_user)
        UserManagedModel.objects.create(id=3, user=other_user)

        for auth in False, True:
            if auth:
                self.client.force_authenticate(self.user)

            tests = [
                (3, 1 if auth else 0, '/usermanagedmodels.json'),
                (3, 3, '/usermanagedmodels/'),
                (3, 3, '/usermanagedmodels.json?page=1'),
                (1, 1, '/usermanagedmodels.json?user_id=%s' % self.user.pk),
                (2, 2, '/usermanagedmodels.json?user_id=%s' % other_user.pk),
            ]
            for expect_count, expect_items, url in tests:
                response = self.client.get(url)
                self.assertTrue(
                    status.is_success(response.status_code), response.data
                )
                self.assertEqual(
                    len(response.data['list']), expect_items,
                    "%s should return %s items for %s" % (
                        url, expect_items,
                        "authed user" if auth else "anonymous user"
                    )
                )
                self.assertEqual(response.data['pages'], 1)
                self.assertEqual(response.data['per_page'], 50)
                self.assertEqual(response.data['count'], expect_count)

    def test_rest_cache_none(self):
        tests = [
            (0, '/items.json'),
            (2, '/items/'),
            (2, '/items.json?page=1'),
        ]
        for expect_items, url in tests:
            response = self.client.get(url)
            self.assertTrue(
                status.is_success(response.status_code), response.data
            )
            self.assertEqual(
                len(response.data['list']), expect_items,
                "%s should return %s items" % (url, expect_items)
            )
            self.assertEqual(response.data['pages'], 1)
            self.assertEqual(response.data['per_page'], 50)
            self.assertEqual(response.data['count'], 2)

    def test_rest_list_head(self):
        response = self.client.head('/')
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.data
        )

    def test_rest_simpleviewset_head(self):
        response = self.client.head('/login')
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.data
        )

    def test_rest_options(self):
        response = self.client.options('/')
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.data
        )

    def test_rest_login_options(self):
        response = self.client.options('/login')
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.data
        )



from django.contrib.auth import authenticate, login, logout
from django.middleware import csrf
from rest_framework.response import Response
from rest_framework.exceptions import AuthenticationFailed
from wq.db import rest
from wq.db.rest.models import get_object_id
from wq.db.rest.views import SimpleViewSet
from .serializers import LoginSerializer


class AuthView(SimpleViewSet):
    def user_info(self, request):
        user_dict = rest.router.serialize(request.user)
        user_dict['id'] = get_object_id(request.user)
        return Response({
            'user': user_dict,
            'config': rest.router.get_config(request.user),
            'csrftoken': csrf.get_token(request),
        })

    def csrf_info(self, request):
        response = {}
        token = csrf.get_token(request)
        if token:
            response['csrftoken'] = token
        return Response(response)


class LoginView(AuthView):
    serializer_class = LoginSerializer

    def list(self, request, *args, **kwargs):
        if request.user.is_authenticated:
            return self.user_info(request)
        else:
            return self.csrf_info(request)

    def create(self, request, *args, **kwargs):
        username = request.POST.get('username')
        password = request.POST.get('password')
        user = authenticate(username=username, password=password)
        if user is not None and user.is_active:
            login(request, user)
            return self.user_info(request)
        else:
            raise AuthenticationFailed(detail="Invalid username or password")


class LogoutView(AuthView):
    def list(self, request, *args, **kwargs):
        if request.user.is_authenticated:
            logout(request)
        return self.csrf_info(request)



from django.core.management.base import BaseCommand
from wq.db import rest
from rest_framework.utils import encoders
import json


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            '--format',
            default='json',
        )
        parser.add_argument(
            '--filename',
            default=None,
        )

    def handle(self, **options):
        if options.get('filename'):
            with open(options['filename'], 'w') as f:
                dump_config(f, **options)
        else:
            dump_config(self.stdout, **options)


def dump_config(f, format='json', **kwargs):
    text = json.dumps(
        rest.router.get_config(),
        cls=encoders.JSONEncoder,
        indent=4,
    )
    if format == "esm":
        text = "const config = %s;\nexport default config;" % text
    elif format == "amd":
        text = "define(%s);" % text
    f.write(text)



from .base import APITestCase
import json

from django.contrib.auth.models import User


class AuthTestCase(APITestCase):
    def setUp(self):
        self.user = User.objects.create(username="testuser")
        self.client.force_authenticate(self.user)

    def test_auth_config_json(self):
        response = self.client.get('/config.json')
        result = json.loads(response.content.decode('utf-8'))
        self.assertTrue("login" in result['pages'])

    def test_auth_login_info(self):
        response = self.client.get('/login.json')
        result = json.loads(response.content.decode('utf-8'))
        self.assertTrue("user" in result)

    def test_auth_context_processors(self):
        response = self.client.get('/auth_context')
        result = response.content.decode('utf-8')
        self.assertHTMLEqual(
            result,
            """
            <div>
                <p>testuser</p>
            </div>
            """
        )



from django.utils.encoding import force_str
from django.urls import re_path
from django.db.utils import DatabaseError

from django.conf import settings
from rest_framework.routers import DefaultRouter, Route
from rest_framework.urlpatterns import format_suffix_patterns
from rest_framework.settings import api_settings
from rest_framework.response import Response

from .model_tools import get_ct
from .permissions import has_perm
from .views import SimpleViewSet, ModelViewSet
from .serializers import ModelSerializer
from .renderers import JSONRenderer, ESMRenderer
from .exceptions import ImproperlyConfigured

import inspect


class ModelRouter(DefaultRouter):
    _models = set()
    _serializers = {}
    _fields = {}
    _querysets = {}
    _filters = {}
    _cache_filters = {}
    _viewsets = {}
    _extra_pages = {}
    _config = {}
    _page_names = {}
    _page_models = {}
    _url_models = {}
    _extra_config = {}

    include_root_view = False
    include_config_view = True
    include_multi_view = True

    default_serializer_class = ModelSerializer

    def __init__(self, trailing_slash=False):
        # Add trailing slash for HTML list views
        self.routes.append(Route(
            url=r'^{prefix}/$',
            mapping={
                'get': 'list',
                'post': 'create'
            },
            name='{basename}-list',
            detail=False,
            initkwargs={'suffix': 'List'}
        ))
        super(ModelRouter, self).__init__(trailing_slash=trailing_slash)

    def register_model(self, model, viewset=None, serializer=None, fields=None,
                       queryset=None, filter=None, cache_filter=None,
                       **kwargs):
        if isinstance(model, str) and '.' in model:
            from django.db.models import get_model
            model = get_model(*model.split('.'))

        if viewset:
            self.register_viewset(model, viewset)
        if queryset is not None:
            self.register_queryset(model, queryset)
        if filter:
            self.register_filter(model, filter)
        if cache_filter:
            self.register_cache_filter(model, cache_filter)
            kwargs.setdefault('cache', 'filter')

        if 'name' not in kwargs:
            kwargs['name'] = model._meta.model_name
        if 'url' not in kwargs:
            url = force_str(model._meta.verbose_name_plural)
            kwargs['url'] = url.replace(' ', '')

        other_model = self._page_models.get(kwargs['name'], None)
        if other_model:
            raise ImproperlyConfigured(
                "Could not register %s: "
                "the name '%s' was already registered for %s"
                % (model, kwargs['name'], other_model)
            )

        other_model = self._url_models.get(kwargs['url'], None)
        if other_model:
            raise ImproperlyConfigured(
                "Could not register %s: "
                "the url '%s' was already registered for %s"
                % (model, kwargs['url'], other_model)
            )

        self.register_config(model, kwargs)

        if serializer:
            self.register_serializer(model, serializer)

        if fields:
            self.register_fields(model, fields)

        self._models.add(model)
        self._page_names[model] = kwargs['name']
        self._page_models[kwargs['name']] = model
        self._url_models[kwargs['url']] = model

    def register_viewset(self, model, viewset):
        self._viewsets[model] = viewset

    def register_serializer(self, model, serializer):
        self._serializers[model] = serializer
        self._base_config = None

    def register_fields(self, model, fields):
        self._fields[model] = fields

    def register_queryset(self, model, queryset):
        self._querysets[model] = queryset

    def register_filter(self, model, filter):
        self._filters[model] = filter

    def register_cache_filter(self, model, cache_filter):
        self._cache_filters[model] = cache_filter

    def register_config(self, model, config):
        for key in ('partial', 'reversed', 'max_local_pages'):
            if key in config:
                raise ImproperlyConfigured(
                    '"%s" is deprecated in favor of "cache"' % key
                )
        self._config[model] = config
        self._base_config = None

    def update_config(self, model, **kwargs):
        if model not in self._config:
            raise RuntimeError("%s must be registered first" % model)
        for key in ('partial', 'reversed', 'max_local_pages'):
            if key in kwargs:
                raise ImproperlyConfigured(
                    '"%s" is deprecated in favor of "cache"' % key
                )
        self._config[model].update(kwargs)
        self._base_config = None

    def set_extra_config(self, **extra):
        self._extra_config.update(extra)
        self._base_config = None

    def get_default_serializer_class(self, model_class):
        return self.default_serializer_class

    def get_class(self, classes, model_class, default=lambda model: None):
        if model_class in classes:
            return classes[model_class]
        else:
            # Make sure we're not dealing with a proxy
            real_model = model_class._meta.concrete_model
            if real_model in classes:
                return classes[real_model]
            else:
                if real_model not in self._config:
                    # FIXME: Probably can just use default instead of erroring
                    raise ImproperlyConfigured(
                        "%s is not registered!" % real_model
                    )
                return default(real_model)

    def get_serializer_for_model(self, model_class, serializer_depth=None):
        serializer = self.get_class(
            self._serializers, model_class,
            self.get_default_serializer_class
        )

        meta = getattr(serializer, 'Meta', object)
        if not getattr(meta, 'model', None):
            if getattr(meta, 'fields', None) or getattr(meta, 'exclude', None):
                include_fields = None
            else:
                include_fields = self._fields.get(model_class, None)
                if not include_fields:
                    raise ImproperlyConfigured(
                        "No serializer fields defined for %s" % model_class
                    )
            serializer = serializer.for_model(
                model_class, include_fields=include_fields
            )

        if serializer_depth is not None:
            if serializer_depth != getattr(meta, 'depth', None):
                serializer = serializer.for_depth(serializer_depth)

        meta = getattr(serializer, 'Meta', object)
        conf = self._config.get(model_class, None)
        if not conf:
            conf = self._config.get(model_class._meta.concrete_model, None)

        if conf:
            wq_conf = getattr(meta, 'wq_config', {}).copy()
            wq_conf.update(conf)
            meta.wq_config = wq_conf
            serializer.Meta = meta

        return serializer

    def serialize(self, obj, many=False, depth=None, request=None):
        if many:
            # assume obj is a queryset
            model = obj.model
            if depth is None:
                depth = 0
        else:
            model = obj.__class__
            if depth is None:
                depth = 1
        serializer = self.get_serializer_for_model(model, depth)
        context = {'router': self}
        if request:
            context['request'] = request
        return serializer(obj, many=many, context=context).data

    def get_paginate_by_for_model(self, model_class):
        config = self.get_model_config(model_class) or {}
        paginate_by = config.get('per_page', None)
        if paginate_by:
            return paginate_by
        return api_settings.PAGE_SIZE

    def paginate(self, model, page_num, request):
        # FIXME: should copy() before modifying but doing so causes recursion
        request.GET = {
            'page': page_num,
        }
        view = self.get_viewset_for_model(model).as_view(
            actions={'get': 'list'},
        )
        return view(request._request).data

    def get_queryset_for_model(self, model, request=None):
        if model in self._page_models:
            model = self._page_models[model]
        if model in self._querysets:
            qs = self._querysets[model]
        else:
            qs = model.objects.all()
        if request and model in self._filters:
            qs = self._filters[model](qs, request)
        return qs

    def get_cache_filter_for_model(self, model):
        return self._cache_filters.get(model, lambda qs, req: qs)

    def get_lookup_for_model(self, model_class):
        config = self.get_model_config(model_class) or {}
        return config.get('lookup', 'pk')

    def get_viewset_for_model(self, model_class):
        if model_class in self._page_models:
            model_class = self._page_models[model_class]
        viewset = self.get_class(
            self._viewsets, model_class, lambda d: ModelViewSet
        )
        lookup = self.get_lookup_for_model(model_class)

        per_page = self.get_paginate_by_for_model(model_class)
        if per_page != api_settings.PAGE_SIZE:
            class CustomPagination(api_settings.DEFAULT_PAGINATION_CLASS):
                page_size = per_page
        else:
            CustomPagination = None

        class ViewSet(viewset):
            model = model_class
            router = self
            lookup_field = lookup

            if CustomPagination:
                pagination_class = CustomPagination

        return ViewSet

    def get_base_url(self):
        return getattr(settings, 'WQ_BASE_URL', '')

    _base_config = None

    @property
    def base_config(self):
        if self._base_config:
            return self._base_config

        from django.contrib.auth.models import AnonymousUser
        user = AnonymousUser()
        pages = {}
        for page in self._extra_pages:
            conf, view = self.get_page(page)
            pages[page] = conf
        for model in self._models:
            try:
                ct = get_ct(model)
            except (RuntimeError, DatabaseError):
                # This can happen before contenttypes is migrated
                ct = str(model._meta)

            if not has_perm(user, ct, 'view'):
                continue

            info = self._config[model].copy()
            info['list'] = True
            for perm in ('add', 'change', 'delete'):
                if has_perm(user, ct, perm):
                    info['can_' + perm] = True

            serializer = self.get_serializer_for_model(model)
            if not hasattr(serializer, 'get_wq_config'):
                raise Exception(model)
            conf = serializer(context={'router': self}).get_wq_config()
            for key in conf:
                if key not in info:
                    info[key] = conf[key]

            pages[info['name']] = info

        site_title = getattr(settings, 'PROJECT_NAME', None)
        base_url = self.get_base_url()

        if not any(conf.get('url') == '' for conf in pages.values()):
            pages.setdefault('index', {
                'url': '',
                'name': 'index',
                'show_in_index': False,
                'verbose_name': site_title
            })

        self._base_config = {
            'pages': pages,
            'site_title': site_title,
            'router': {'base_url': base_url},
            'store': {
                'service': base_url,
                'defaults': {'format': 'json'},
            },
        }
        if getattr(settings, 'WQ_CONFIG', None):
            self._base_config.update(settings.WQ_CONFIG)
        if settings.DEBUG:
            self._base_config['debug'] = True
        self._base_config.update(self._extra_config)
        return self._base_config

    def get_config(self, user=None):
        if user is None or not user.is_authenticated:
            return self.base_config

        # Add user-specific permissions to configuration
        config = {
            key: val.copy() if hasattr(val, 'copy') else val
            for key, val in self.base_config.items()
        }
        for page, info in config['pages'].items():
            if not info.get('list', False):
                continue
            try:
                ct = get_ct(self._page_models[page])
            except (RuntimeError, DatabaseError):
                continue
            info = info.copy()
            for perm in ('add', 'change', 'delete'):
                if has_perm(user, ct, perm):
                    info['can_' + perm] = True
            config['pages'][page] = info

        return config

    def add_page(self, name, config, view=None):
        if view is None:
            class PageView(SimpleViewSet):
                template_name = name + '.html'

                def list(self, request, *args, **kwargs):
                    return Response(config)
            view = PageView
        if 'name' not in config:
            config['name'] = name
        if 'url' not in config:
            config['url'] = name
        self._extra_pages[name] = config, view
        self._base_config = None

    def get_page(self, page):
        return self._extra_pages[page]

    def get_page_config(self, name, user=None):
        config = self.get_config(user)
        return config['pages'].get(name, None)

    def get_model_config(self, model, user=None):
        if model in self._page_models:
            model = self._page_models[model]

        # First, check models registered with API
        if model in self._page_names:
            config = self.get_config(user)
            return config['pages'][self._page_names[model]]

        # Then check config cache directly (in case model was configured but
        # not fully registered as part of API)
        if model in self._config:
            return self._config[model]
        return None

    def model_is_registered(self, model):
        return model in self._models

    def get_config_view(self):
        class ConfigView(SimpleViewSet):
            renderer_classes = [JSONRenderer, ESMRenderer]

            def list(this, request, *args, **kwargs):
                if request.accepted_renderer.format == 'js':
                    config = self.base_config
                else:
                    config = self.get_config(request.user)
                return Response(config)

        return ConfigView

    def get_index(self, user):
        config = self.get_config(user)

        def page_sort(page):
            is_list = page.get('list', False)
            return (not is_list, page['name'])

        pages = sorted(config['pages'].values(), key=page_sort)
        return {
            'pages': pages
        }

    def get_index_view(self):
        class IndexView(SimpleViewSet):
            def list(this, request, *args, **kwargs):
                return Response(self.get_index(request.user))
        return IndexView

    def get_multi_view(self):
        class MultipleListView(SimpleViewSet):
            def list(this, request, *args, **kwargs):
                urls = request.GET.get('lists', '').split(',')
                return self.get_multi(request, urls)
        return MultipleListView

    def get_multi(self, request, urls):
        conf = self.get_config(request.user)
        conf_by_url = {
            conf['url']: (page, conf)
            for page, conf
            in self.get_config(request.user)['pages'].items()
        }
        result = {}
        for listurl in urls:
            if listurl not in conf_by_url:
                continue
            page, conf = conf_by_url[listurl]
            model = self._page_models[page]
            result[listurl] = self.paginate(model, 1, request)
        return Response(result)

    def get_urls(self):
        # Register viewsets with DefaultRouter just before returning urls

        # The root viewset (with a url of "") should be registered last
        root = {}

        def register(config, viewset):
            if config['url'] == "":
                root['view'] = viewset
                root['name'] = config['name']
            else:
                self.register(config['url'], viewset, config['name'])

        # List (model) views
        for model in self._models:
            viewset = self.get_viewset_for_model(model)
            config = self._config[model]
            register(config, viewset)

        # Extra/custom viewsets
        for name in self._extra_pages:
            register(*self._extra_pages[name])

        if self.include_config_view:
            # /config.js
            self.register('config', self.get_config_view(), 'config')

        if self.include_multi_view:
            # /multi.json
            self.register('multi', self.get_multi_view(), 'multi')

        if not root:
            # default index
            self.register('', self.get_index_view(), 'index')

        urls = super(ModelRouter, self).get_urls()

        if root:
            # / - Skip registration and directly generate custom URLs
            urls.extend(self.get_root_urls(root['view'], root['name']))

        return urls

    def get_root_urls(self, viewset, name):
        lookup = self.get_lookup_regex(viewset)
        routes = self.get_routes(viewset)
        urls = []
        for route in routes:
            pattern = route.url.format(
                prefix='',
                lookup=lookup,
                trailing_slash=self.trailing_slash,
            )

            # Remove leading slash from detail view URLs
            pattern = pattern.replace('/', '', 1)

            mapping = self.get_method_map(viewset, route.mapping)
            if not mapping:
                continue
            view = viewset.as_view(mapping, **route.initkwargs)
            urls.append(re_path(pattern, view, name=name))

        return format_suffix_patterns(urls)

    def get_routes(self, viewset):
        routes = super(ModelRouter, self).get_routes(viewset)
        model = getattr(viewset, "model", None)
        if not model:
            return routes

        # Custom routes

        # Re-register list view, with an additional keyword to filter this
        # model by parent models (e.g. foreign keys)

        # /[parentmodel_url]/[foreignkey_value]/[model_url]
        try:
            ct = get_ct(model)
        except (RuntimeError, DatabaseError):
            # This can happen before contenttypes is migrated
            return routes
        for pct, fields in ct.get_foreign_keys().items():
            if not pct.is_registered():
                continue
            if len(fields) > 1:
                # Multiple foreign keys to same parent model; can't
                # automatically determine which one to use
                continue
            if pct.urlbase == '':
                purlbase = ''
            else:
                purlbase = pct.urlbase + '/'
            routes.append(Route(
                url=(
                    '^' + purlbase + r'(?P<' + fields[0] +
                    r'>[^\/\?]+)/{prefix}{trailing_slash}$'
                ),
                mapping={'get': 'list'},
                name="{basename}-for-%s" % fields[0],
                detail=False,
                initkwargs={'suffix': 'List'},
            ))

        for cct in ct.get_children():
            if not cct.is_registered():
                continue
            cbase = cct.urlbase
            routes.append(Route(
                url='^%s-by-{prefix}' % cbase,
                mapping={'get': 'list'},
                name="%s-by-%s" % (cct.identifier, ct.identifier),
                detail=False,
                initkwargs={'target': cbase, 'suffix': 'List'},
            ))

        if hasattr(viewset, 'extra_routes'):
            routes += viewset.extra_routes()

        return routes

    @property
    def version(self):
        if not hasattr(self, '_version'):
            vtxt = getattr(settings, 'VERSION_TXT', None)
            if vtxt is None:
                self._version = None
            else:
                vfile = open(vtxt, 'r')
                self._version = vfile.read()
                vfile.close()
        return self._version

    @property
    def urls(self):
        urls = super(ModelRouter, self).urls
        try:
            # FIXME: Remove in 2.0
            caller = inspect.stack()[1]
            code_context = getattr(caller, 'code_context', caller[4])
            if 'include' in code_context[0]:
                return urls, 'wq'
        except Exception:
            pass
        return urls, 'wq', 'wq'


# Default router instance, c.f. django.contrib.admin.sites.site
router = ModelRouter()



from django.contrib.contenttypes.models import (
    ContentType as DjangoContentType,
    ContentTypeManager as DjangoContentTypeManager
)
from django.utils.encoding import force_str
from .model_tools import get_ct, get_object_id, get_by_identifier


__all__ = (
    'ContentType',
    'get_ct',
    'get_object_id',
    'get_by_identifier',
)


class ContentTypeManager(DjangoContentTypeManager):
    def get_by_identifier(self, identifier):
        # TODO: What if model is configured with a different name?
        return self.get(model=identifier)


class ContentType(DjangoContentType):
    objects = ContentTypeManager()

    @property
    def identifier(self):
        from . import router
        conf = router._config.get(self.model_class(), {})
        return conf.get('name', self.model)

    @property
    def urlbase(self):
        cls = self.model_class()
        if cls is None:
            return None
        config = self.get_config()
        if config:
            return config['url']
        urlbase = force_str(cls._meta.verbose_name_plural)
        return urlbase.replace(' ', '')

    # Get foreign keys for this content type
    def get_foreign_keys(self):
        cls = self.model_class()
        if cls is None:
            return []
        parents = {}
        for f in cls._meta.fields:
            rel = f.remote_field
            if rel is not None and type(rel).__name__ == 'ManyToOneRel':
                parent = get_ct(rel.model)
                parents.setdefault(parent, [])
                parents[parent].append(f.name)
        return parents

    def get_parents(self):
        return set(self.get_foreign_keys().keys())

    def get_children(self, include_rels=False):
        cls = self.model_class()
        if cls is None:
            return []

        rels = [
            field for field in cls._meta.get_fields()
            if (field.one_to_many or field.one_to_one)
            and field.auto_created and not field.concrete
        ]

        children = [(get_ct(rel.related_model), rel) for rel in rels]
        if include_rels:
            return set(children)
        else:
            return set(child[0] for child in children)

    def get_config(self, user=None):
        from . import router  # avoid circular import
        cls = self.model_class()
        return router.get_model_config(cls, user)

    def is_registered(self):
        from . import router  # avoid circular import
        cls = self.model_class()
        return router.model_is_registered(cls)

    class Meta:
        proxy = True



from django.utils.module_loading import autodiscover_modules
from .routers import ModelRouter, router
from .management.commands.dump_config import dump_config
from django.conf import settings


__all__ = (
    "ModelRouter",
    "router",
    "autodiscover",
    "default_app_config",
)


def autodiscover():
    autodiscover_modules('rest', register_to=None)

    if getattr(settings, 'WQ_CONFIG_FILE', None):
        with open(settings.WQ_CONFIG_FILE, 'w') as f:
            dump_config(f, format='esm')


default_app_config = 'wq.db.rest.apps.RestConfig'



from rest_framework import serializers
from wq.db.rest.serializers import ModelSerializer


class LoginSerializer(serializers.Serializer):
    username = serializers.CharField()
    password = serializers.CharField(style={'input_type': 'password'})


class UserSerializer(ModelSerializer):
    class Meta:
        exclude = ('id', 'password', 'user_permissions', 'groups',)
        read_only_fields = ('last_login', 'date_joined')



from .base import APITestCase
from rest_framework import status
from django.contrib.auth.models import User
import json
from tests.naturalkey_app.models import ModelWithNaturalKey


class NaturalKeyTestCase(APITestCase):
    def setUp(self):
        self.user = User.objects.create(username='testuser', is_superuser=True)
        self.client.force_authenticate(user=self.user)

    def test_naturalkey_config(self):
        response = self.client.get('/config.json')
        self.assertEqual([
            {
                'name': 'note',
                'label': 'Note',
                'type': 'text',
                'bind': {'required': True},
            },
            {
                'name': 'key',
                'label': 'Key',
                'type': 'group',
                'bind': {'required': True},
                'children': [{
                     'name': 'parent[slug]',
                     'label': 'Parent',
                     'type': 'select one',
                     'wq:ForeignKey': 'naturalkeyparent',
                     'wq:length': 50,
                     'bind': {'required': True},
                }, {
                     'name': 'date',
                     'label': 'Date',
                     'type': 'date',
                     'bind': {'required': True},
                }]
            }
        ], response.data['pages']['modelwithnaturalkey']['form'])

    def test_naturalkey_post(self):
        form = {
            'key[parent][slug]': 'test-key',
            'key[date]': '2016-12-31',
            'note': 'Test Note',
        }

        response = self.client.post('/modelwithnaturalkeys.json', form)
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.data
        )
        instance = ModelWithNaturalKey.objects.get(pk=response.data['id'])

        self.assertEqual(json.loads(response.content.decode('utf-8')), {
            'id': instance.id,
            'key_id': instance.key_id,

            'note': 'Test Note',
            'key': {
                'parent': {'slug': 'test-key'},
                'date': '2016-12-31',
            },

            'label': 'test-key on 2016-12-31: Test Note',
            'key_label': 'test-key on 2016-12-31',
        })



from rest_framework.negotiation import DefaultContentNegotiation


JSON = 'application/json'
GEOJSON = 'application/geo+json'


class ContentNegotiation(DefaultContentNegotiation):
    """
    Some clients (such as Mapbox GL) might only Accept: application/json
    """
    def filter_renderers(self, renderers, format):
        self._format = format
        return super().filter_renderers(renderers, format)

    def get_accept_list(self, request):
        accepts = super().get_accept_list(request)
        if getattr(self, '_format', None) == 'geojson':
            if JSON in accepts and GEOJSON not in accepts:
                accepts.append(GEOJSON)
        return accepts



# Generated by Django 2.1.7 on 2019-03-21 07:55

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Attribute',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=10)),
                ('is_active', models.BooleanField()),
                ('category', models.CharField(blank=True, max_length=10)),
            ],
        ),
        migrations.CreateModel(
            name='Campaign',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
        ),
        migrations.CreateModel(
            name='CustomAttachment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=10)),
            ],
        ),
        migrations.CreateModel(
            name='CustomPatternModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=10)),
            ],
        ),
        migrations.CreateModel(
            name='CustomType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=10)),
            ],
        ),
        migrations.CreateModel(
            name='CustomTypedAttachment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(blank=True, max_length=10, null=True)),
                ('value', models.FloatField(blank=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='CustomTypedPatternModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=10)),
            ],
        ),
        migrations.CreateModel(
            name='Entity',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('campaign', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='patterns_app.Campaign')),
            ],
            options={
                'verbose_name_plural': 'entities',
            },
        ),
        migrations.CreateModel(
            name='FilterableModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=10)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='IdentifiedModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(blank=True, db_index=True, max_length=255)),
                ('slug', models.CharField(blank=True, max_length=255, unique=True)),
            ],
            options={
                'abstract': False,
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='Value',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('attribute', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='patterns_app.Attribute')),
                ('entity', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='values', to='patterns_app.Entity')),
            ],
        ),
        migrations.AddField(
            model_name='filterablemodel',
            name='parent',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='patterns_app.IdentifiedModel'),
        ),
        migrations.AddField(
            model_name='customtypedattachment',
            name='parent',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='attachments', to='patterns_app.CustomTypedPatternModel'),
        ),
        migrations.AddField(
            model_name='customtypedattachment',
            name='type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='patterns_app.CustomType'),
        ),
        migrations.AddField(
            model_name='customattachment',
            name='parent',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='attachments', to='patterns_app.CustomPatternModel'),
        ),
        migrations.AddField(
            model_name='attribute',
            name='campaign',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='patterns_app.Campaign'),
        ),
    ]



try:
    from django.urls import path
except ImportError:
    from django.conf.urls import url
    path = None

from wq.db import rest
from tests.patterns_app.views import FilterableView
from .settings import WITH_NONROOT


if WITH_NONROOT:
    base_url = 'wqsite/'
else:
    base_url = ''


if path:
    urlpatterns = [
        path(base_url + 'filterable/<path:ids>', FilterableView.as_view()),
        path(base_url, rest.router.urls),
    ]
else:
    # FIXME: Remove in 2.0
    urlpatterns = [
        url(
            r'^' + base_url + 'filterable/(?P<ids>.+)',
            FilterableView.as_view()
        ),
        url(r'^' + base_url, rest.router.urls),
    ]



import os
import sys
from setuptools import setup, find_packages

LONG_DESCRIPTION = """
Django REST API for offline mobile surveys and field data collection.
"""


def readme():
    # Attempt to load markdown file
    try:
        readme = open('README.md')
    except IOError:
        return LONG_DESCRIPTION
    return readme.read()


def create_wq_namespace():
    """
    Generate the wq namespace package
    (not checked in, as it technically is the parent of this folder)
    """
    if os.path.isdir("wq"):
        return
    os.makedirs("wq")
    init = open(os.path.join("wq", "__init__.py"), 'w')
    init.write("__import__('pkg_resources').declare_namespace(__name__)")
    init.close()


def find_wq_packages(submodule):
    """
    Add submodule prefix to found packages, since the packages within each wq
    submodule exist at the top level of their respective repositories.
    """
    packages = ['wq', submodule]
    package_dir = {submodule: '.'}
    for package in find_packages():
        if package.split('.')[0] in ('wq', 'tests'):
            continue
        full_name = submodule + "." + package
        packages.append(full_name)
        package_dir[full_name] = package.replace('.', os.sep)

    return packages, package_dir


create_wq_namespace()
packages, package_dir = find_wq_packages('wq.db')

setup(
    name='wq.db',
    use_scm_version=True,
    author='S. Andrew Sheppard',
    author_email='andrew@wq.io',
    url='https://wq.io/wq.db',
    license='MIT',
    packages=packages,
    package_dir=package_dir,
    namespace_packages=['wq'],
    entry_points={'wq': 'wq.db=wq.db'},
    description=LONG_DESCRIPTION.strip(),
    long_description=readme(),
    long_description_content_type="text/markdown",
    install_requires=[
        'Django>=1.11,<5.0',
        'djangorestframework>=3.8.0,<4.0',
        'html-json-forms',
        'natural-keys>=1.6.0',
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Framework :: Django',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Framework :: Django :: 3.1',
        'Framework :: Django :: 3.2',
        'Framework :: Django :: 4.0',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Text Processing :: Markup :: HTML',
        'Topic :: Scientific/Engineering :: GIS',
        'Topic :: Database :: Database Engines/Servers',
    ],
    test_suite='tests',
    setup_requires=[
        'setuptools_scm',
    ],
    project_urls={
        'Homepage': 'https://wq.io/wq.db/',
        'Documentation': 'https://wq.io/',
        'Source': 'https://github.com/wq/wq.db',
        'Release Notes': 'https://github.com/wq/wq.db/releases',
        'Issues': 'https://github.com/wq/wq.db/issues',
        'CI': 'https://github.com/wq/wq.db/actions/workflows/test.yml',
    },
)
